@@ -33,7 +33,9 @@ from gettext import gettext as _
 import gettext
 import os
 import subprocess
+import multiprocessing
 import logging
+import traceback
 import threading
 import sys
 import time
@@ -107,6 +109,29 @@ POCKET_SECURITY = "security"
 POCKET_UPDATES = "updates"
 POCKET_BACKPORTS = "backports"
 
+SIGTERM = 15
+
+# Overload multiprocessing.Process.run() so we can catch run's exception.
+# https://stackoverflow.com/a/63773140
+class Process(multiprocessing.Process):
+    def __init__(self, *args, **kwargs):
+        multiprocessing.Process.__init__(self, *args, **kwargs)
+        self._pconn, self._cconn = multiprocessing.Pipe()
+        self._exception = None
+    def run(self):
+        try:
+            multiprocessing.Process.run(self)
+            self._cconn.send(None)
+        except Exception:
+            tb = traceback.format_exc()
+            self._cconn.send(tb)
+    @property
+    def exception(self):
+        if self._pconn.poll():
+            self._exception = self._pconn.recv()
+        if not (self._exception is None):
+            return Exception(self._exception)
+
 
 class UpdateSubscriptions(object):
     security = False
@@ -235,6 +260,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.init_apt_cache()
 
         self.pk_task = None
+        self.search_process = None
 
         # Put some life into the user interface:
         self.init_auto_update()
@@ -282,11 +308,16 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         # On the additional drivers page, don't show the backend revert button.
         if page == self.vbox_drivers:
             self.button_revert.set_visible(False)
-            if not self.detect_called:
-                GLib.idle_add(lambda: threading.Thread(target=self.detect_drivers).start())
+            self.maybe_detect_drivers()
         else:
             self.button_revert.set_visible(True)
 
+    def maybe_detect_drivers(self):
+        if not self.search_process or not self.search_process.is_alive():
+            t = threading.Thread(target=self.detect_drivers, daemon=True)
+            self.label_driver_detail.set_text(_("Searching for available drivers..."))
+            t.start()
+
     def update_interface(self):
         """ abstract interface to keep the UI alive """
         while Gtk.events_pending():
@@ -1380,18 +1411,11 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.box_driver_action.pack_end(self.button_driver_restart, False, False, 0)
         self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
 
-        props = {"halign":Gtk.Align.CENTER, "valign":Gtk.Align.CENTER,
-                 "vexpand":True, "visible":True}
-        self.label_driver_detail = Gtk.Label(label=_("Searching for available "
-                                                     "drivers..."), **props)
-        self.box_driver_detail.add(self.label_driver_detail)
-
         self.progress_bar = Gtk.ProgressBar()
         self.box_driver_action.pack_end(self.progress_bar, False, False, 0)
         self.progress_bar.set_visible(False)
 
         self.devices = {}
-        self.detect_called = False
         self.driver_changes = []
         self.orig_selection = {}
         # HACK: the case where the selection is actually "Do not use"; is a little
@@ -1403,22 +1427,67 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.nonfree_drivers = 0
         self.ui_building = False
 
+    def on_detect_drivers_cancel_retry(self, button):
+        if self.search_process.is_alive():
+            self.search_process.terminate()
+        else:
+            self.maybe_detect_drivers()
+        self.update_drivers_control_button()
+
+    def update_drivers_control_button(self):
+        if self.search_process.is_alive():
+            self.button_drivers_control.set_label(_("Cancel"))
+        else:
+            self.button_drivers_control.set_label(_("Retry"))
+
+    def wrapper_system_device_drivers(self, queue, cache):
+        devices = detect.system_device_drivers(cache)
+        queue.put_nowait(devices)
+
     def detect_drivers(self):
         # WARNING: This is run in a separate thread.
-        self.detect_called = True
+        queue = multiprocessing.Queue()
+        self.search_process = Process(
+            target=self.wrapper_system_device_drivers,
+            args=(queue, self.apt_cache,),
+            daemon=True
+        )
+        self.search_process.start()
+        GLib.idle_add(lambda: self.update_drivers_control_button())
+        error = None
         try:
-            self.devices = detect.system_device_drivers(self.apt_cache)
-        except:
+            self.search_process.join()
+            if self.search_process.exception:
+                raise(self.search_process.exception)
+            if self.search_process.exitcode == 0:
+                self.devices = queue.get_nowait()
+            # In the multiprocessing module, if it was terminated by signal N,
+            # the exit code will be the negative value -N.
+            elif self.search_process.exitcode == -SIGTERM:
+                error = _("The search for drivers was cancelled.")
+            else:
+                raise Exception(f"Exit code: {self.search_process.exitcode}")
+        except Exception as e:
             # Catch all exceptions and feed them to apport.
-            GLib.idle_add(self.label_driver_detail.set_text,
-                          _("An error occurred while searching for drivers."))
+            error = _("An error occurred while searching for drivers.")
+            logging.error("An error occurred while searching for drivers."
+                          "Exception: %s", e)
 
             # For apport to catch this exception. See
             # http://bugs.python.org/issue1230540
             sys.excepthook(*sys.exc_info())
+
+        # Using an Idle to ensure the function is called in the main (UI) thread
+        GLib.idle_add(lambda: self.on_detect_drivers_done(error))
+
+    def on_detect_drivers_done(self, err=None):
+        if err is None:
+            self.button_drivers_control.set_sensitive(False)
+            self.show_drivers()
             return
 
-        GLib.idle_add(self.show_drivers)
+        self.label_driver_detail.set_text(err)
+        self.update_drivers_control_button()
 
     def on_driver_selection_changed(self, button, modalias, pkg_name=None):
         if self.ui_building:
@@ -1575,6 +1644,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         return (overall_status, icon, returned_drivers)
 
     def show_drivers(self):
+        self.box_driver_detail.remove(self.buttonbox_driver_detail)
         if not self.devices:
             # No drivers found.
             self.label_driver_detail.set_text(_("No additional drivers available."))
